#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2013 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import print_function, absolute_import

import os
import ssl
import socket
import unittest
import pyuv

from unittest import SkipTest

from gruvi.ssl import SslPipe, SslTransport
from gruvi.sslcompat import SSLContext
from support import UnitTest
from test_transports import EventLoopTest, TransportTest


def communicate(buf, client, server, clientssl, serverssl):
    """Send *buf* from *client* to *server* over SSL.

    The *clientssl* and *serverssl* arguments are potentially empty list of
    initial SSL data. The clientssl list is SSL data from the client to send to
    the server, the serverssl list is SSL data to send from the server to the
    client.

    The data that is received on the server is returned.
    """
    received = []
    offset = bytes_received = 0
    while True:
        ssldata, appdata = server.feed_ssldata(b''.join(clientssl))
        serverssl += ssldata
        received += appdata
        for chunk in appdata:
            if not chunk:
                break  # close_notify
            bytes_received += len(chunk)
        clientssl = []
        ssldata, appdata = client.feed_ssldata(b''.join(serverssl))
        clientssl += ssldata
        for chunk in appdata:
            assert len(chunk) == 0
        serverssl = []
        ssldata, offset = client.feed_appdata(buf, offset)
        clientssl += ssldata
        if not clientssl and not serverssl:
            break
    received = b''.join(received)
    return received


class TestSslPipe(UnitTest):
    """Test suite for the SslPipe class."""

    def setUp(self):
        if not os.access(self.certname, os.R_OK):
            raise SkipTest('no certificate available')
        super(TestSslPipe, self).setUp()
        context = SSLContext(ssl.PROTOCOL_SSLv23)
        context.load_cert_chain(self.certname, self.certname)
        self.client = SslPipe(context, False)
        self.server = SslPipe(context, True)

    def test_wrapped(self):
        # Send a simple chunk of data over SSL from client to server.
        buf = b'x' * 1000
        client, server = self.client, self.server
        # The client starts the handshake.
        clientssl = client.do_handshake()
        self.assertEqual(len(clientssl), 1)
        self.assertGreater(len(clientssl[0]), 0)
        # The server waits.
        serverssl = server.do_handshake()
        self.assertEqual(len(serverssl), 0)
        received = communicate(buf, client, server, clientssl, serverssl)
        self.assertTrue(client.wrapped)
        self.assertTrue(server.wrapped)
        self.assertEqual(received, buf)

    def test_echo(self):
        # Send a chunk from client to server and echo it back.
        buf = b'x' * 1000
        client, server = self.client, self.server
        clientssl = client.do_handshake()
        serverssl = server.do_handshake()
        received = communicate(buf, client, server, clientssl, serverssl)
        self.assertEqual(received, buf)
        received = communicate(buf, server, client, [], [])
        self.assertEqual(received, buf)

    def test_shutdown(self):
        # Test a clean shutdown of the SSL protocol.
        client, server = self.client, self.server
        buf = b'x' * 1000
        clientssl = client.do_handshake()
        serverssl = server.do_handshake()
        received = communicate(buf, client, server, clientssl, serverssl)
        self.assertEqual(received, buf)
        # The client initiates a shutdown.
        clientssl = client.shutdown()
        self.assertEqual(len(clientssl), 1)
        self.assertGreater(len(clientssl[0]), 0)  # the c->s close_notify alert
        # Communicate the close_notify to the server.
        serverssl, appdata = server.feed_ssldata(clientssl[0])
        # b'' means close notify: acknowledge it
        self.assertEqual(serverssl, [])
        self.assertEqual(appdata, [b''])
        serverssl = server.shutdown()
        # Now we should have a close_notify
        self.assertEqual(len(serverssl), 1)
        self.assertGreater(len(serverssl[0]), 0)  # the s->c close_notify
        self.assertEqual(appdata, [b''])
        self.assertFalse(server.wrapped)
        # Send back the server response to the client.
        clientssl, appdata = client.feed_ssldata(serverssl[0])
        self.assertEqual(clientssl, [])
        self.assertEqual(appdata, [b''])
        self.assertFalse(client.wrapped)

    def test_unwrapped(self):
        # Send data over an unencrypted channel.
        client, server = self.client, self.server
        buf = b'x' * 1000
        received = communicate(buf, client, server, [], [])
        self.assertEqual(received, buf)

    def test_unwrapped_after_wrapped(self):
        # Send data over SSL, then unwrap, and send data in the clear.
        client, server = self.client, self.server
        buf = b'x' * 1000
        # Send some data in the clear.
        received = communicate(buf, client, server, [], [])
        self.assertEqual(received, buf)
        # Now start the handshake and send some encrypted data.
        clientssl = client.do_handshake()
        server.do_handshake()
        received = communicate(buf, client, server, clientssl, [])
        self.assertTrue(client.wrapped)
        self.assertTrue(server.wrapped)
        self.assertEqual(received, buf)
        # Move back to clear again.
        clientssl = client.shutdown()
        received = communicate(buf, client, server, clientssl, [])
        self.assertEqual(received, b'')
        serverssl = server.shutdown()
        received = communicate(buf, client, server, [], serverssl)
        self.assertFalse(client.wrapped)
        self.assertFalse(server.wrapped)
        self.assertEqual(received, buf)
        # And back to encrypted again..
        clientssl = client.do_handshake()
        server.do_handshake()
        received = communicate(buf, client, server, clientssl, [])
        self.assertTrue(client.wrapped)
        self.assertTrue(server.wrapped)
        self.assertEqual(received, buf)

    def test_simultaneous_shutdown(self):
        # Test a simultaenous shutdown.
        client, server = self.client, self.server
        buf = b'x' * 1000
        # Start an encrypted session.
        clientssl = client.do_handshake()
        server.do_handshake()
        received = communicate(buf, client, server, clientssl, [])
        self.assertEqual(received, buf)
        # Tear it down concurrently.
        clientssl = client.shutdown()
        serverssl = server.shutdown()
        received = communicate(buf, client, server, clientssl, serverssl)
        self.assertFalse(client.wrapped)
        self.assertFalse(server.wrapped)
        self.assertEqual(received, buf)  # this was sent in the clear


class SslTransportTest(TransportTest):

    def setUp(self):
        if not os.access(self.certname, os.R_OK):
            raise SkipTest('no certificate available')
        super(SslTransportTest, self).setUp()
        self.context = SSLContext(ssl.PROTOCOL_SSLv3)
        self.context.load_cert_chain(self.certname, self.certname)

    def create_transport(self, handle, protocol, server_side):
        transport = SslTransport(handle, self.context, server_side)
        transport.start(protocol)
        return transport


class TestSslTcpTransport(SslTransportTest, EventLoopTest):

    def create_handle(self):
        return pyuv.TCP(self.loop)

    def bind_handle(self, handle):
        host = socket.gethostbyname('localhost')
        handle.bind((host, 0))
        return handle.getsockname()


class TestSslPipeTransport(SslTransportTest, EventLoopTest):

    def create_handle(self):
        return pyuv.Pipe(self.loop)

    def bind_handle(self, handle):
        addr = self.pipename('test-pipe')
        handle.bind(addr)
        return addr


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import socket
import functools
import pyuv
import six
import errno

from . import logging
from .hub import get_hub, switchpoint, switch_back
from .sync import Event
from .errors import Timeout
from .transports import TransportError, Transport
from .ssl import SslTransport, create_ssl_context
from .address import getaddrinfo, saddr

__all__ = ['create_connection', 'create_server', 'Endpoint', 'Client', 'Server']


def _use_af_unix(addr):
    """Return whether to open a :class:`pyuv.Pipe` via an AF_UNIX socket."""
    # This is used on Linux only to support abstract sockets.
    if isinstance(addr, six.text_type) and u'\x00' not in addr \
            or isinstance(addr, six.binary_type) and b'\x00' not in addr:
        return False
    return sys.platform in ('linux', 'linux2', 'linux3')

def _af_unix_helper(handle, address, op):
    """Connect or bind a :class:`pyuv.Pipe` to an AF_UNIX socket.

    We use this on Linux to work around the limitation in the libuv API that
    socket names cannot have null bytes in them (required for abstract
    sockets on Linux).
    """
    # Note that on Linux, connect() to an abstract socket never returns EAGAIN.
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.setblocking(False)
    try:
        if op == 'connect':
            sock.connect(address)
        elif op == 'bind':
            sock.bind(address)
        fd = os.dup(sock.fileno())
    except IOError as e:
        # Connecting to an AF_UNIX socket never gives EAGAIN on Linux.
        assert e.errno != errno.EAGAIN
        # Convert from Unix errno -> libuv errno via the symbolic error name
        errname = 'UV_{}'.format(errno.errocode.get(e.errno, 'UNKNOWN'))
        errnum = getattr(pyuv.errno, errname, pyuv.errno.UV_UNKNOWN)
        raise pyuv.error.PipeError(errnum, os.strerror(e.errno))
    finally:
        sock.close()
    handle.open(fd)


@switchpoint
def create_connection(protocol_factory, address, ssl=False, ssl_args={},
                      family=0, flags=0, local_address=None, timeout=None, mode='rw'):
    """Create a new client connection.

    This method creates a new :class:`pyuv.Handle`, connects it to *address*,
    and then waits for the connection to be established. When the connection is
    established, the handle is wrapped in a transport, and a new protocol
    instance is created by calling *protocol_factory*. The protocol is then
    connected to the transport by calling the transport's
    :meth:`~BaseTransport.start` method which in turn calls
    :meth:`~BaseProtocol.connection_made` on the protocol. Finally the results
    are returned as a ``(transport, protocol)`` tuple.

    The address may be either be a string, a (host, port) tuple, a
    ``pyuv.Stream`` handle or a file descriptor:

    * If the address is a string, this method connects to a named pipe using a
      :class:`pyuv.Pipe` handle.
    * If the address is a tuple, this method connects to a TCP/IP service using
      a :class:`pyuv.TCP` handle. The host element of the tuple the IP address
      or DNS name, and the port element is the port number or service name. The
      tuple is always passed to :func:`getaddrinfo` for resolution together
      with the *family* and *flags* arguments.
    * If the address is a ``pyuv.Stream`` instance, it must be an already
      connected stream.
    * If the address is a file descriptor, then it is attached to a
      :class:`pyuv.TTY` stream if :func:`os.isatty` returns true, or to a
      :class:`pyuv.Pipe` instance otherwise.

    The *ssl* parameter indicates whether SSL should be used on top of the
    stream transport. If an SSL connection is desired, then *ssl* can be set to
    ``True`` or to an :class:`ssl.SSLContext` instance. In the former case a
    default SSL context is created. In the case of Python 2.x the :mod:`ssl`
    module does not define an SSL context object and you can use
    :func:`create_ssl_context` instead which works across all supported Python
    versions. The *ssl_args* argument may be used to pass keyword arguments to
    :class:`SslTransport`.

    If an SSL connection was selected, the resulting transport will be a
    :class:`SslTransport` instance, otherwise it will be a :class:`Transport`
    instance.

    The *mode* parameter specifies if the transport should be put in read-only
    (``'r'``), write-only (``'w'``) or read-write (``'rw'``) mode. For TTY
    transports, the mode must be either read-only or write-only. For all other
    transport the mode should usually be read-write.

    The *local_address* keyword argument is relevant only for TCP transports.
    If provided, it specifies the local address to bind to.
    """
    hub = get_hub()
    log = logging.get_logger()
    if isinstance(address, (six.binary_type, six.text_type)):
        handle_type = pyuv.Pipe
        addresses = [address]
    elif isinstance(address, tuple):
        handle_type = pyuv.TCP
        result = getaddrinfo(address[0], address[1], family, socket.SOCK_STREAM,
                             socket.IPPROTO_TCP, flags)
        addresses = [res[4] for res in result]
    elif isinstance(address, int):
        if os.isatty(address):
            if mode not in ('r', 'w'):
                raise ValueError("mode: must be either 'r' or 'w' for tty")
            handle = pyuv.TTY(hub.loop, address, mode == 'r')
        else:
            handle = pyuv.Pipe(hub.loop, True)
            handle.open(address)
        addresses = []; error = None
    elif isinstance(address, pyuv.Stream):
        handle = address
        addresses = []; error = None
    else:
        raise TypeError('expecting a string, tuple, fd, or pyuv.Stream')
    for addr in addresses:
        log.debug('trying address {}', saddr(addr))
        handle = handle_type(hub.loop)
        try:
            error = None
            if handle_type is pyuv.Pipe and _use_af_unix(addr):
                _af_unix_helper(handle, addr, 'connect')
            else:
                with switch_back(timeout) as switcher:
                    handle.connect(addr, switcher)
                    result = hub.switch()
                    _, error = result[0]
        except pyuv.error.UVError as e:
            error = e[0]
        except Timeout:
            error = pyuv.errno.UV_ETIMEDOUT
        if not error:
            break
        handle.close()
        log.warning('connect() failed with error {}', error)
    if error:
        log.warning('all addresses failed')
        raise TransportError.from_errno(error)
    if local_address:
        handle.bind(*local_address)
    protocol = protocol_factory()
    protocol._timeout = timeout
    if ssl:
        context = ssl if hasattr(ssl, 'set_ciphers') else create_ssl_context()
        transport = SslTransport(handle, context, False, **ssl_args)
    else:
        transport = Transport(handle, mode)
    events = transport.start(protocol)
    if events:
        for event in events:
            event.wait()
    return (transport, protocol)


@switchpoint
def create_server(protocol_factory, address=None, ssl=False, ssl_args={},
                  family=0, flags=0, backlog=128):
    """
    Create a new network server.

    This creates one or more :class:`pyuv.Handle` instances bound to *address*,
    puts them in listen mode and starts accepting new connections. For each
    accepted connection, a new transport is created which is connected to a new
    protocol instance obtained by calling *protocol_factory*.

    The *address* argument may be either be a string, a ``(host, port)`` tuple,
    or a ``pyuv.Stream`` handle:

    * If the address is a string, this method creates a new :class:`pyuv.Pipe`
      instance and binds it to *address*.
    * If the address is a tuple, this method creates one or more
      :class:`pyuv.TCP` handles. The host element of the tuple the IP address
      or DNS name, and the port element is the port number or service name. The
      tuple is passed to :func:`getaddrinfo` for resolution together with the
      *family* and *flags* arguments. A transport is created for each resolved
      address.
    * If the address is a ``pyuv.Stream`` handle, it must already be bound to
      an address.

    The *ssl* parameter indicates whether SSL should be used for accepted
    connections. See :func:`create_connection` for a description of the *ssl*
    and *ssl_args* parameters.

    The *backlog* parameter specifies the listen backlog i.e the maximum
    number of not yet accepted connections to queue.

    The return value is a :class:`Server` instance that can be used to control
    the listening transports.
    """
    server = Server(protocol_factory)
    server.listen(address, ssl=ssl, ssl_args=ssl_args, family=family,
                  flags=flags, backlog=backlog)
    return server


class Endpoint(object):
    """A communications endpoint."""

    def __init__(self, protocol_factory, timeout=None):
        """
        The *protocol_factory* argument constructs a new protocol instance.
        Normally you would pass a :class:`~gruvi.Protocol` subclass.

        The *timeout* argument specifies a timeout for various network operations.
        """
        self._protocol_factory = protocol_factory
        self._timeout = timeout
        self._hub = get_hub()
        self._log = logging.get_logger(self)

    @property
    def timeout(self):
        """The network timeout."""
        return self._timeout

    def close(self):
        """Close the endpoint."""
        raise NotImplementedError


class Client(Endpoint):
    """A client endpoint."""

    def __init__(self, protocol_factory, timeout=None):
        super(Client, self).__init__(protocol_factory, timeout=timeout)
        self._transport = None
        self._protocol = None

    @property
    def transport(self):
        """Return the transport, or ``None`` if not connected."""
        return self._transport

    @property
    def protocol(self):
        """Return the protocol, or ``None`` if not connected."""
        return self._protocol

    @switchpoint
    def connect(self, address, **kwargs):
        """Connect to *address* and wait for the connection to be established.

        See :func:`~gruvi.create_connection` for a description of *address*
        and the supported keyword arguments.
        """
        if self._transport:
            raise RuntimeError('already connected')
        kwargs.setdefault('timeout', self._timeout)
        conn = create_connection(self._protocol_factory, address, **kwargs)
        self._transport = conn[0]
        self._transport._log = self._log
        self._protocol = conn[1]
        self._protocol._log = self._log
        self._protocol._timeout = self._timeout

    @switchpoint
    def close(self):
        """Close the connection."""
        if self._transport is None:
            return
        self._transport.close()
        self._transport._closed.wait()
        self._transport = None
        self._protocol = None


class Server(Endpoint):
    """A server endpoint."""

    max_connections = None

    def __init__(self, protocol_factory, timeout=None):
        super(Server, self).__init__(protocol_factory, timeout=timeout)
        self._handles = []
        self._addresses = []
        self._connections = dict()
        self._all_closed = Event()
        self._all_closed.set()

    @property
    def addresses(self):
        """The addresses this server is listening on."""
        return self._addresses

    @property
    def connections(self):
        """An iterator yielding the (transport, protocol) pairs for each connection."""
        return self._connections.items()

    def _on_new_connection(self, ssl, ssl_args, handle, error):
        # Callback used with handle.listen().
        assert handle in self._handles
        if error:
            self._log.warning('error {} in listen() callback', error)
            return
        client = type(handle)(self._hub.loop)
        handle.accept(client)
        if self.max_connections is not None and len(self._connections) >= self.max_connections:
            self._log.warning('max connections reached, dropping new connection')
            client.close()
            return
        if ssl:
            context = ssl if hasattr(ssl, 'set_ciphers') else create_ssl_context()
            transport = SslTransport(client, context, True, **ssl_args)
        else:
            transport = Transport(client)
        transport._log = self._log
        transport._server = self
        self._all_closed.clear()
        if __debug__:
            self._log.debug('new connection on {}', saddr(client.getsockname()))
            if hasattr(client, 'getpeername'):
                self._log.debug('remote peer is {}', saddr(client.getpeername()))
        protocol = self._protocol_factory()
        protocol._log = self._log
        protocol._timeout = self._timeout
        self._connections[transport] = protocol
        self.connection_made(transport, protocol)
        transport.start(protocol)

    def _on_close_complete(self, transport, protocol, exc=None):
        # Called by Transport._on_close_complete
        self._connections.pop(transport, None)
        if not self._connections:
            self._all_closed.set()
        self.connection_lost(transport, protocol, exc)

    def connection_made(self, transport, protocol):
        """Called when a new connection is made."""

    def connection_lost(self, transport, protocol, exc=None):
        """Called when a connection is lost."""

    @switchpoint
    def listen(self, address, ssl=False, ssl_args={}, family=0, flags=0, backlog=128):
        """Create a new transport, bind it to *address*, and start listening
        for new connections.

        See :func:`create_server` for a description of *address* and the
        supported keyword arguments.
        """
        handles = []
        if isinstance(address, six.string_types):
            handle_type = pyuv.Pipe
            addresses = [address]
        elif isinstance(address, tuple):
            handle_type = pyuv.TCP
            result = getaddrinfo(address[0], address[1], family, socket.SOCK_STREAM,
                                 socket.IPPROTO_TCP, flags)
            addresses = [res[4] for res in result]
        elif isinstance(address, pyuv.Stream):
            handles.append(address)
            addresses = []
        else:
            raise TypeError('expecting a string, tuple or pyuv.Stream')
        for addr in addresses:
            handle = handle_type(self._hub.loop)
            try:
                if handle_type is pyuv.Pipe and _use_af_unix(addr):
                    _af_unix_helper(handle, addr, 'bind')
                else:
                    handle.bind(addr)
            except pyuv.error.UVError as e:
                self._log.warning('bind error {!r}, skipping {}', e[0], saddr(addr))
                continue
            handles.append(handle)
        addresses = []
        for handle in handles:
            callback = functools.partial(self._on_new_connection, ssl, ssl_args)
            handle.listen(callback, backlog)
            addr = handle.getsockname()
            self._log.debug('listen on {}', saddr(addr))
            addresses.append(addr)
        self._handles += handles
        self._addresses += addresses

    @switchpoint
    def close(self):
        """Close the listening sockets and all accepted connections."""
        for handle in self._handles:
            if not handle.closed:
                handle.close()
        del self._handles[:]
        for transport, _ in self.connections:
            transport.close()
        self._all_closed.wait()

    @switchpoint
    def run(self):
        """Run the event loop and start serving requests.

        This method is useful in scripts that run only one server instance. In
        more complicated applications you normally call :meth:`Hub.switch`
        explicitly.
        """
        get_hub().switch()



#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys

from setuptools import setup, Extension


version_info = {
    'name': 'gruvi',
    'version': '0.10.3',
    'description': 'Pythonic async IO with libuv and fibers',
    'author': 'Geert Jansen',
    'author_email': 'geertj@gmail.com',
    'url': 'https://github.com/geertj/gruvi',
    'license': 'MIT',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
}

topdir, _ = os.path.split(os.path.abspath(__file__))


def get_requirements():
    """Parse a requirements.txt file and return as a list."""
    with open(os.path.join(topdir, 'requirements.txt')) as fin:
        lines = fin.readlines()
    lines = [line.strip() for line in lines]
    return lines


def main():
    os.chdir(topdir)
    ext_modules = []
    # Note that on Windows it's more involved to compile _sslcompat because
    # there's no system provided OpenSSL and you need to match the version that
    # was used to compile your Python.
    if sys.version_info[:2] < (3, 5):
        ext_modules.append(Extension('_sslcompat', ['src/sslcompat.c'],
                                     libraries=['ssl', 'crypto']))
    setup(
        packages=['gruvi', 'gruvi_vendor.txdbus'],
        package_dir={'': 'lib', 'gruvi_vendor': 'vendor'},
        setup_requires=['cffi >= 1.0.0'],
        install_requires=get_requirements(),
        cffi_modules=['src/build_http.py:ffi', 'src/build_jsonrpc.py:ffi'],
        ext_package='gruvi',
        ext_modules=ext_modules,
        zip_safe=False,
        **version_info
    )


if __name__ == '__main__':
    main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import pyuv
import contextlib
import socket
import struct

from . import logging, compat
from .util import docfrom
from .errors import Error
from .sync import Event

__all__ = ['TransportError', 'BaseTransport', 'Transport', 'DatagramTransport']


class TransportError(Error):
    """A transport error."""

    def __init__(self, message, errno=None):
        super(TransportError, self).__init__(message)
        self._errno = errno

    @property
    def errno(self):
        return self._errno

    @classmethod
    def from_errno(cls, errno):
        """Create a new instance from a :mod:`pyuv.errno` error code."""
        message = '{}: {}'.format(pyuv.errno.errorcode.get(errno, errno),
                                    pyuv.errno.strerror(errno))
        return cls(message, errno)


class BaseTransport(object):
    """Base class for :mod:`pyuv` based transports. There is no public
    constructor."""

    default_write_buffer = 65536

    def __init__(self, handle, mode):
        self._handle = handle
        self._readable = 'r' in mode
        self._writable = 'w' in mode
        self._protocol = None
        self._server = None
        self._log = logging.get_logger()
        self._write_buffer_size = 0
        self._write_buffer_high = self.default_write_buffer
        self._write_buffer_low = self.default_write_buffer // 2
        self._closing = False
        self._error = None
        self._reading = False
        self._writing = False
        self._started = False
        self._can_write = Event()
        self._closed = Event()

    def start(self, protocol):
        """Bind to *protocol* and start calling callbacks on it. """
        if self._protocol is not None:
            raise TransportError('already started')
        self._protocol = protocol
        self._protocol.connection_made(self)
        if self._readable:
            self.resume_reading()
        if self._writable:
            self._writing = True
            self._can_write.set()

    def _check_status(self):
        # Check the status of the transport.
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._protocol is None:
            raise TransportError('transport not started')
        elif self._handle.closed:
            raise TransportError('transport was closed')

    def get_write_buffer_size(self):
        """Return the total number of bytes in the write buffer."""
        return self._write_buffer_size

    def get_write_buffer_limits(self):
        """Return the write buffer limits as a ``(low, high)`` tuple."""
        return self._write_buffer_low, self._write_buffer_high

    def set_write_buffer_limits(self, high=None, low=None):
        """Set the low and high watermark for the write buffer."""
        if high is None:
            high = self.write_buffer_size
        if low is None:
            low = high // 2
        if low > high:
            low = high
        self._write_buffer_high = high
        self._write_buffer_low = low

    def _maybe_resume_protocol(self):
        # Called after the write buffer size decreased. Possibly resume the protocol.
        if self._closing or self._handle.closed or self._writing:
            return
        if self.get_write_buffer_size() <= self.get_write_buffer_limits()[0]:
            self._writing = True
            self._can_write.set()
            self._protocol.resume_writing()

    def _maybe_pause_protocol(self):
        # Called after the write buffer size increased. Possibly pause the protocol.
        if self._closing or self._handle.closed or not self._writing:
            return
        if self.get_write_buffer_size() > self.get_write_buffer_limits()[0]:
            self._writing = False
            self._can_write.clear()
            self._protocol.pause_writing()

    def resume_reading(self):
        """Resume calling callbacks on the protocol.

        As a relaxation from the requirements in Python asyncio, this method
        may be called even if reading was already paused, and also if a close
        is pending. This simplifies protocol design, especially in combination
        with SSL where reading may need to be enabled by the transport itself,
        without knowledge of the protocol, to complete handshakes.
        """
        raise NotImplementedError

    def pause_reading(self):
        """Pause calling callbacks on the protocol.

        This method may be called even if reading was already paused or a close
        is pending. See the note in :meth:`resume_reading`.
        """
        raise NotImplementedError

    def _maybe_close(self):
        # Check a pending close request and close.
        if not self._closing or self._write_buffer_size > 0:
            return
        if not self._handle.closed:
            self._handle.close(self._on_close_complete)
            assert self._handle.closed
        self._closing = False

    def _on_close_complete(self, handle):
        # Callback used with handle.close().
        assert handle is self._handle
        assert handle.closed
        if self._server is not None:
            self._server._on_close_complete(self, self._protocol, self._error)
        self._protocol.connection_lost(self._error)
        self._protocol = None  # remove cycle to help garbage collection
        self._closed.set()

    def close(self):
        """Close the transport after all oustanding data has been written."""
        if self._closing or self._handle.closed:
            return
        elif self._protocol is None:
            raise TransportError('transport not started')
        # If the write buffer is empty, close now. Otherwise defer to
        # _on_write_complete that will close when the buffer is empty.
        if self._write_buffer_size == 0:
            self._handle.close(self._on_close_complete)
            assert self._handle.closed
        else:
            self._closing = True

    def abort(self):
        """Close the transport immediately."""
        if self._handle.closed:
            return
        elif self._protocol is None:
            raise TransportError('transport not started')
        self._handle.close(self._on_close_complete)
        assert self._handle.closed

    def get_extra_info(self, name, default=None):
        """Get transport specific data.

        The following information is available for all transports:

        ==============  =================================================
        Name            Description
        ==============  =================================================
        ``'handle'``    The pyuv handle that is being wrapped.
        ==============  =================================================
        """
        if name == 'handle':
            return self._handle
        else:
            return default


class Transport(BaseTransport):
    """A connection oriented transport."""

    def __init__(self, handle, mode='rw'):
        """
        The *handle* argument is the pyuv handle for which to create the
        transport. It must be a ``pyuv.Stream`` instance, so either a
        :class:`pyuv.TCP`, :class:`pyuv.Pipe` or a :class:`pyuv.TTY`.

        The *mode* argument specifies if this is transport is read-only
        (``'r'``), write-only (``'w'``) or read-write (``'rw'``).
        """
        if not isinstance(handle, pyuv.Stream):
            raise TypeError("handle: expecting a 'pyuv.Stream' instance, got {!r}"
                                .format(type(handle).__name__))
        super(Transport, self).__init__(handle, mode)

    @docfrom(BaseTransport.get_write_buffer_size)
    def get_write_buffer_size(self):
        # Return the size of the write buffer. Return the actual number of
        # outstanding bytes, which libuv keeps track of for us. Note that we
        # _also_ use self._write_buffer_size to keep track of the total number
        # of oustanding write requests.. This allows us to keep track of e.g.
        # write_eof() that doesn't write actual bytes.
        return self._handle.write_queue_size

    def _on_read_complete(self, handle, data, error):
        # Callback used with handle.start_read().
        assert handle is self._handle
        if self._error:
            self._log.warning('ignore read status {} after error', error)
        elif error == pyuv.errno.UV_EOF:
            if self._protocol.eof_received():
                self._log.debug('EOF received, protocol wants to continue')
            else:
                self._log.debug('EOF received, closing transport')
                self.close()
        elif error:
            self._log.warning('pyuv error {} in read callback', error)
            self._error = TransportError.from_errno(error)
            self.abort()
        elif data:
            self._protocol.data_received(data)

    @docfrom(BaseTransport.resume_reading)
    def resume_reading(self):
        # Resume reading
        self._check_status()
        if not self._readable:
            raise TransportError('transport is not readable')
        if not self._reading:
            self._handle.start_read(self._on_read_complete)
            self._reading = True

    @docfrom(BaseTransport.pause_reading)
    def pause_reading(self):
        # Pause reading
        self._check_status()
        if not self._readable:
            raise TransportError('transport is not readable')
        if self._reading:
            self._handle.stop_read()
            self._reading = False

    def _on_write_complete(self, handle, error):
        # Callback used with handle.write().
        assert handle is self._handle
        self._write_buffer_size -= 1
        assert self._write_buffer_size >= 0
        if self._error:
            self._log.debug('ignore write status {} after error', error)
        # UV_ECANCELED happens when a handle is closed that has a write backlog.
        # This happens when Transport.abort() is called.
        elif error and error != pyuv.errno.UV_ECANCELED:
            self._log.warning('pyuv error {} in write callback', error)
            self._error = TransportError.from_errno(error)
            self.abort()
        self._maybe_resume_protocol()
        self._maybe_close()

    def write(self, data):
        """Write *data* to the transport."""
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError("data: expecting a bytes-like instance, got {!r}"
                                .format(type(data).__name__))
        self._check_status()
        if not self._writable:
            raise TransportError('transport is not writable')
        if self._closing:
            raise TransportError('transport is closing')
        try:
            self._handle.write(data, self._on_write_complete)
        except pyuv.error.UVError as e:
            self._error = TransportError.from_errno(e.args[0])
            self.abort()
            raise compat.saved_exc(self._error)
        # We only keep track of the number of outstanding write requests
        # outselves. See note in get_write_buffer_size().
        self._write_buffer_size += 1
        self._maybe_pause_protocol()

    def writelines(self, seq):
        """Write all elements from *seq* to the transport."""
        for line in seq:
            self.write(line)

    def write_eof(self):
        """Shut down the write direction of the transport."""
        self._check_status()
        if not self._writable:
            raise TransportError('transport is not writable')
        if self._closing:
            raise TransportError('transport is closing')
        try:
            self._handle.shutdown(self._on_write_complete)
        except pyuv.error.UVError as e:
            self._error = TransportError.from_errno(e.args[0])
            self.abort()
            raise compat.saved_exc(self._error)
        self._write_buffer_size += 1

    def can_write_eof(self):
        """Whether this transport can close the write direction."""
        return True

    def get_extra_info(self, name, default=None):
        """Get transport specific data.

        In addition to the fields from :meth:`BaseTransport.get_extra_info`,
        the following information is also available:

        ==================  ===================================================
        Name                Description
        ==================  ===================================================
        ``'sockname'``      The socket name i.e. the result of the
                            ``getsockname()`` system call.
        ``'peername'``      The peer name i.e. the result of the
                            ``getpeername()`` system call.
        ``'winsize'``       The terminal window size as a ``(cols, rows)``
                            tuple. Only available for :class:`pyuv.TTY`
                            handles.
        ``'unix_creds'``    The Unix credentials of the peer as a
                            ``(pid, uid, gid)`` tuple. Only available for
                            :class:`pyuv.Pipe` handles on Unix.
        ==================  ===================================================
        """
        if name == 'sockname':
            if not hasattr(self._handle, 'getsockname'):
                return default
            try:
                return self._handle.getsockname()
            except pyuv.error.UVError:
                return default
        elif name == 'peername':
            if not hasattr(self._handle, 'getpeername'):
                return default
            try:
                return self._handle.getpeername()
            except pyuv.error.UVError:
                return default
        elif name == 'winsize':
            if not hasattr(self._handle, 'get_winsize'):
                return default
            try:
                return self._handle.get_winsize()
            except pyuv.error.UVError:
                return default
        elif name == 'unix_creds':
            # In case you're wondering, DBUS needs this.
            if not isinstance(self._handle, pyuv.Pipe) or not hasattr(socket, 'SO_PEERCRED'):
                return default
            try:
                fd = self._handle.fileno()
                sock = socket.fromfd(fd, socket.AF_UNIX, socket.SOCK_DGRAM)  # will dup()
                with contextlib.closing(sock):
                    creds = sock.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED,
                                            struct.calcsize('3i'))
            except socket.error:
                return default
            return struct.unpack('3i', creds)
        else:
            return super(Transport, self).get_extra_info(name, default)


class DatagramTransport(BaseTransport):
    """A datagram transport."""

    def __init__(self, handle, mode='rw'):
        """
        The *handle* argument is the pyuv handle for which to create the
        transport. It must be a :class:`pyuv.UDP` instance.

        The *mode* argument specifies if this is transport is read-only
        (``'r'``), write-only (``'w'``) or read-write (``'rw'``).
        """
        if not isinstance(handle, pyuv.UDP):
            raise TypeError("handle: expecting a 'pyuv.UDP' instance, got {!r}"
                                .format(type(handle).__name__))
        super(DatagramTransport, self).__init__(handle, mode)

    @docfrom(BaseTransport.get_write_buffer_size)
    def get_write_buffer_size(self):
        # Return the size of the write buffer. See the note for the same method
        # in Transport.
        # This was recently added to pyuv, so don't put a hard dependency on
        # it. UDP buffering in the write direction very OS dependent anyway.
        return getattr(self._handle, 'send_queue_size', 0)

    def _on_recv_complete(self, handle, addr, flags, data, error):
        """Callback used with handle.start_recv()."""
        assert handle is self._handle
        if error:
            self._log.warning('pyuv error {} in recv callback', error)
            self._protocol.error_received(TransportError.from_errno(error))
        elif flags:
            assert flags & pyuv.UV_UDP_PARTIAL
            self._log.warning('ignoring partial datagram')
        elif data:
            self._protocol.datagram_received(data, addr)

    @docfrom(BaseTransport.resume_reading)
    def resume_reading(self):
        # Resume reading
        self._check_status()
        if not self._readable:
            raise TransportError('transport is not readable')
        if not self._reading:
            self._handle.start_recv(self._on_recv_complete)
            self._reading = True

    @docfrom(BaseTransport.pause_reading)
    def pause_reading(self):
        # Pause reading
        self._check_status()
        if not self._readable:
            raise TransportError('transport is not readable')
        if self._reading:
            self._handle.stop_recv()
            self._reading = False

    def _on_send_complete(self, handle, error):
        """Callback used with handle.send()."""
        assert handle is self._handle
        self._write_buffer_size -= 1
        assert self._write_buffer_size >= 0
        if self._error:
            self._log.debug('ignore sendto status {} after error', error)
        # See note in _on_write_complete() about UV_ECANCELED
        elif error and error != pyuv.errno.UV_ECANCELED:
            self._log.warning('pyuv error {} in sendto callback', error)
            self._protocol.error_received(TransportError.from_errno(error))
        self._maybe_resume_protocol()
        self._maybe_close()

    def sendto(self, data, addr=None):
        """Send a datagram containing *data* to *addr*.

        The *addr* argument may be omitted only if the handle was bound to a
        default remote address.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError("data: expecting a bytes-like instance, got {!r}"
                                .format(type(data).__name__))
        self._check_status()
        if not self._writable:
            raise TransportError('transport is not writable')
        try:
            self._handle.send(addr, data, self._on_send_complete)
        except pyuv.error.UVError as e:
            error = TransportError.from_errno(e.args[0])
            # Try to discern between permanent and transient errors. Permanent
            # errors close the transport. This list is very likely not complete.
            if error.errno != pyuv.errno.UV_EBADF:
                raise error
            self._error = error
            self.abort()
        self._write_buffer_size += 1
        self._maybe_pause_protocol()



# # This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import random
import unittest

from gruvi.callbacks import dllist, Node
from gruvi.callbacks import check as check_dllist
from gruvi.callbacks import add_callback, remove_callback, pop_callback
from gruvi.callbacks import walk_callbacks, run_callbacks
from gruvi.callbacks import has_callback, clear_callbacks
from support import UnitTest


def get_callbacks(obj):
    callbacks = []
    while True:
        callback = pop_callback(obj)
        if not callback:
            return
        callbacks.append(callback)
    return callbacks


class TestDllist(UnitTest):

    def test_basic(self):
        dll = dllist()
        check_dllist(dll)
        self.assertIsNone(dll.first)
        self.assertIsNone(dll.last)
        self.assertEqual(len(dll), 0)
        # insert first element
        n1 = Node('foo')
        dll.insert(n1)
        self.assertIs(dll.first, n1)
        self.assertIs(dll.last, n1)
        self.assertEqual(len(dll), 1)
        self.assertIn(n1, dll)
        check_dllist(dll)
        # insert second at end
        n2 = Node('bar')
        dll.insert(n2)
        self.assertIs(dll.first, n1)
        self.assertIs(dll.last, n2)
        self.assertEqual(len(dll), 2)
        self.assertIn(n2, dll)
        check_dllist(dll)
        # insert in middle
        n3 = Node('baz')
        dll.insert(n3, before=n2)
        self.assertIs(dll.first, n1)
        self.assertIs(dll.last, n2)
        self.assertEqual(len(dll), 3)
        self.assertIn(n3, dll)
        check_dllist(dll)
        # remove middle
        dll.remove(n3)
        self.assertIs(dll.first, n1)
        self.assertIs(dll.last, n2)
        self.assertEqual(len(dll), 2)
        self.assertNotIn(n3, dll)
        check_dllist(dll)
        # remove first
        dll.remove(n1)
        self.assertIs(dll.first, n2)
        self.assertIs(dll.last, n2)
        self.assertEqual(len(dll), 1)
        self.assertNotIn(n1, dll)
        check_dllist(dll)
        # remove remaining element
        dll.remove(n2)
        self.assertIsNone(dll.first)
        self.assertIsNone(dll.last)
        self.assertEqual(len(dll), 0)
        self.assertNotIn(n2, dll)
        check_dllist(dll)

    def test_iter(self):
        dll = dllist()
        for i in range(10):
            dll.insert(Node(10+i))
            check_dllist(dll)
        value = 10
        for node in dll:
            self.assertIsInstance(node, Node)
            self.assertEqual(node.callback, value)
            value += 1
        check_dllist(dll)

    def test_args(self):
        dll = dllist()
        def callback():
            pass
        n1 = Node(callback)
        self.assertEqual(n1.callback, callback)
        self.assertIsNone(n1.args)
        dll.insert(n1)
        n2 = Node(callback, 'foo')
        self.assertIs(n2.callback, callback)
        self.assertEqual(n2.args, 'foo')
        dll.insert(n2)
        nodes = list(dll)
        self.assertEqual(nodes[0].callback, callback)
        self.assertIsNone(nodes[0].args)
        self.assertEqual(nodes[1].callback, callback)
        self.assertEqual(nodes[1].args, 'foo')

    def test_clear(self):
        dll = dllist()
        for i in range(10):
            dll.insert(Node(i))
        self.assertEqual(len(dll), 10)
        dll.clear()
        self.assertEqual(len(dll), 0)

    def test_many_nodes(self):
        nodes = []
        dll = dllist()
        count = 10000
        for i in range(count):
            before = random.choice(nodes) if nodes else None
            node = Node(i)
            dll.insert(node, before)
            nodes.append(node)
            self.assertEqual(len(dll), i+1)
            if i % 100 == 0:
                check_dllist(dll)
        check_dllist(dll)
        for i in range(count):
            r = random.randint(0, len(nodes)-1)
            node = nodes[r]; del nodes[r]
            dll.remove(node)
            self.assertEqual(len(dll), count-i-1)
            if i % 100 == 0:
                check_dllist(dll)
        check_dllist(dll)


class Object(object):

    def __init__(self):
        self._callbacks = None


class TestCallbacks(UnitTest):

    def test_add_pop_one(self):
        obj = Object()
        def callback(*args):
            pass
        add_callback(obj, callback)
        self.assertIsInstance(obj._callbacks, Node)
        self.assertIs(obj._callbacks.callback, callback)
        self.assertEqual(obj._callbacks.args, ())
        self.assertEqual(pop_callback(obj), (callback, ()))
        self.assertEqual(obj._callbacks, None)
        add_callback(obj, callback, (1, 2))
        self.assertIsInstance(obj._callbacks, Node)
        self.assertIs(obj._callbacks.callback, callback)
        self.assertEqual(obj._callbacks.args, (1, 2))
        self.assertEqual(pop_callback(obj), (callback, (1, 2)))
        self.assertEqual(obj._callbacks, None)

    def test_add_pop_multiple(self):
        obj = Object()
        def callback(*args):
            pass
        for i in range(5):
            add_callback(obj, callback)
        self.assertIsInstance(obj._callbacks, dllist)
        for i in range(5):
            self.assertEqual(len(obj._callbacks), 5-i)
            self.assertEqual(pop_callback(obj), (callback, ()))
        self.assertIsNone(obj._callbacks)
        for i in range(5):
            add_callback(obj, callback, (1, 2))
        self.assertIsInstance(obj._callbacks, dllist)
        for i in range(5):
            self.assertEqual(len(obj._callbacks), 5-i)
            self.assertEqual(pop_callback(obj), (callback, (1, 2)))
        self.assertIsNone(obj._callbacks)

    def test_add_remove_one(self):
        obj = Object()
        def callback(*args):
            pass
        handle = add_callback(obj, callback)
        self.assertIsInstance(obj._callbacks, Node)
        remove_callback(obj, handle)
        self.assertIs(obj._callbacks, None)

    def test_add_remove_multiple(self):
        obj = Object()
        def callback(*args):
            pass
        handles = []
        for i in range(5):
            handles.append(add_callback(obj, callback))
        self.assertIsInstance(obj._callbacks, dllist)
        for i in range(5):
            self.assertEqual(len(obj._callbacks), 5-i)
            remove_callback(obj, handles[i])
        self.assertIsNone(obj._callbacks)
        handles = []
        for i in range(5):
            handles.append(add_callback(obj, callback, (1, 2)))
        self.assertIsInstance(obj._callbacks, dllist)
        for i in range(5):
            self.assertEqual(len(obj._callbacks), 5-i)
            remove_callback(obj, handles[i])
        self.assertIsNone(obj._callbacks)

    def test_has_callback(self):
        obj = Object()
        def callback():
            pass
        n1 = add_callback(obj, callback)
        self.assertTrue(has_callback(obj, n1))
        n2 = add_callback(obj, callback)
        self.assertTrue(has_callback(obj, n2))
        remove_callback(obj, n2)
        self.assertFalse(has_callback(obj, n2))
        remove_callback(obj, n1)
        self.assertFalse(has_callback(obj, n1))

    def test_clear_callbacks(self):
        obj = Object()
        def callback():
            pass
        for i in range(10):
            add_callback(obj, callback)
        self.assertEqual(len(obj._callbacks), 10)
        clear_callbacks(obj)
        self.assertIsNone(obj._callbacks)

    def test_run_one(self):
        obj = Object()
        cbargs = []
        def callback(*args):
            cbargs.append(args)
        add_callback(obj, callback, (obj,))
        run_callbacks(obj)
        self.assertEqual(cbargs, [(obj,)])
        self.assertIsNone(obj._callbacks)

    def test_run_multiple(self):
        obj = Object()
        cbargs = []
        def callback(*args):
            cbargs.append(args)
        for i in range(5):
            add_callback(obj, callback, (obj,))
        run_callbacks(obj)
        self.assertEqual(cbargs, [(obj,)]*5)

    def test_walk_one(self):
        obj = Object()
        walk = []
        result = True
        def callback():
            pass
        add_callback(obj, callback, ('foo',))
        def walker(callback, args):
            walk.append((callback, args))
            return result
        walk_callbacks(obj, walker)
        self.assertEqual(walk, [(callback, ('foo',))])
        self.assertIsInstance(obj._callbacks, Node)
        result = False
        walk_callbacks(obj, walker)
        self.assertEqual(walk, [(callback, ('foo',))]*2)
        self.assertIsNone(obj._callbacks)

    def test_walk_multiple(self):
        obj = Object()
        def callback():
            pass
        ref = [(callback, i) for i in range(5)]
        for r in ref:
            add_callback(obj, r[0], r[1])
        walk = []
        result = True
        def walker(callback, args):
            walk.append((callback, args))
            return result
        walk_callbacks(obj, walker)
        self.assertEqual(walk, ref)
        self.assertEqual(len(obj._callbacks), 5)
        walk = []
        result = False
        walk_callbacks(obj, walker)
        self.assertEqual(walk, ref)
        self.assertFalse(obj._callbacks)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import unittest

from gruvi import http
from gruvi.http import HttpServer, HttpClient, HttpMessage, HttpProtocol, ParsedUrl
from gruvi.http import parse_content_type, parse_te, parse_trailer, parse_url
from gruvi.http import get_header, remove_headers
from gruvi.stream import Stream, StreamClient
from gruvi.sync import Queue

from support import UnitTest, MockTransport

URL = ParsedUrl


class TestParseContentType(UnitTest):

    def test_simple(self):
        parsed = parse_content_type('text/plain')
        self.assertEqual(parsed, ('text/plain', {}))

    def test_params(self):
        parsed = parse_content_type('text/plain; charset=foo')
        self.assertEqual(parsed[0], 'text/plain')
        self.assertEqual(parsed[1], {'charset': 'foo'})

    def test_iso8859_1(self):
        parsed = parse_content_type('text/plain; foo="bar\xfe"')
        self.assertEqual(parsed, ('text/plain', {'foo': 'bar\xfe'}))

    def test_param_whitespace(self):
        parsed = parse_content_type('text/plain;   charset=foo    ')
        self.assertEqual(parsed, ('text/plain', {'charset': 'foo'}))

    def test_param_quoted(self):
        parsed = parse_content_type('text/plain; charset="foo bar"')
        self.assertEqual(parsed, ('text/plain', {'charset': 'foo bar'}))

    def test_param_quoted_pair(self):
        parsed = parse_content_type('text/plain; charset="foo\\"bar"')
        self.assertEqual(parsed, ('text/plain', {'charset': 'foo"bar'}))

    def test_param_empty(self):
        parsed = parse_content_type('text/plain; charset=""')
        self.assertEqual(parsed, ('text/plain', {'charset': ''}))

    def test_param_multiple(self):
        parsed = parse_content_type('text/plain; foo=bar; baz=qux')
        self.assertEqual(parsed, ('text/plain', {'foo': 'bar', 'baz': 'qux'}))

    def test_param_multiple_missing_semi(self):
        parsed = parse_content_type('text/plain; foo=bar baz=qux')
        self.assertEqual(parsed, ('text/plain', {'foo': 'bar'}))


class TestParseTE(UnitTest):

    def test_simple(self):
        parsed = parse_te('chunked')
        self.assertEqual(parsed, [('chunked', None)])

    def test_multiple(self):
        parsed = parse_te('chunked, deflate')
        self.assertEqual(parsed, [('chunked', None), ('deflate', None)])

    def test_qvalue(self):
        parsed = parse_te('deflate; q=0.5')
        self.assertEqual(parsed, [('deflate', '0.5')])

    def test_case_insensitive(self):
        parsed = parse_te('dEfLaTe; Q=0.5')
        self.assertEqual(parsed, [('dEfLaTe', '0.5')])

    def test_illegal_qvalue(self):
        parsed = parse_te('deflate; q=2.5')
        self.assertEqual(parsed, [('deflate', None)])

    def test_multiple_qvalue(self):
        parsed = parse_te('deflate; q=0.5, zlib; q=0.8')
        self.assertEqual(parsed, [('deflate', '0.5'), ('zlib', '0.8')])


class TestParseTrailer(UnitTest):

    def test_simple(self):
        parsed = parse_trailer('foo')
        self.assertEqual(parsed, ['foo'])

    def test_multiple(self):
        parsed = parse_trailer('foo, bar')
        self.assertEqual(parsed, ['foo', 'bar'])

    def test_spacing(self):
        parsed = parse_trailer('foo   , bar   ')
        self.assertEqual(parsed, ['foo', 'bar'])

    def test_wrong_separator(self):
        parsed = parse_trailer('foo; bar')
        self.assertEqual(parsed, ['foo'])


class TestParseUrl(UnitTest):

    def test_test(self):
        self.assertEqual(URL(), ('', '', '', '', '', '', ''))
        self.assertEqual(URL(scheme='http'), ('http', '', '', '', '', '', ''))
        self.assertEqual(URL(host='foo'), ('', 'foo', '', '', '', '', ''))
        self.assertEqual(URL(path='/path'), ('', '', '/path', '', '', '', ''))
        self.assertEqual(URL(query='foo=bar'), ('', '', '', 'foo=bar', '', '', ''))
        self.assertEqual(URL(fragment='baz'), ('', '', '', '', 'baz', '', ''))
        self.assertEqual(URL(port='80'), ('', '', '', '', '', '80', ''))
        self.assertEqual(URL(userinfo='user:pass'), ('', '', '', '', '', '', 'user:pass'))

    def test_origin(self):
        parsed = parse_url('/path')
        self.assertEqual(parsed, URL(path='/path'))

    def test_absolute(self):
        parsed = parse_url('http://example.com/path')
        self.assertEqual(parsed, URL('http', 'example.com', '/path'))

    def test_authority(self):
        parsed = parse_url('example.com:80', is_connect=True)
        self.assertEqual(parsed, URL('', 'example.com', port='80'))

    def test_authority_error(self):
        self.assertRaises(ValueError, parse_url, '/path', is_connect=True)
        self.assertRaises(ValueError, parse_url, 'http://example.com:80', is_connect=True)
        self.assertRaises(ValueError, parse_url, '*', is_connect=True)

    def test_asterisk(self):
        parsed = parse_url('*')
        self.assertEqual(parsed, URL(path='*'))

    def test_userinfo(self):
        parsed = parse_url('http://user:pass@example.com')
        self.assertEqual(parsed, URL('http', 'example.com', userinfo='user:pass'))

    def test_port(self):
        parsed = parse_url('http://example.com:80')
        self.assertEqual(parsed, URL('http', 'example.com', port='80'))

    def test_userinfo_port(self):
        parsed = parse_url('http://user:pass@example.com:80')
        self.assertEqual(parsed, URL('http', 'example.com', port='80', userinfo='user:pass'))

    def test_default_scheme(self):
        parsed = parse_url('www.example.com')
        self.assertEqual(parsed, URL('http', 'www.example.com'))
        parsed = parse_url('http://www.example.com')
        self.assertEqual(parsed, URL('http', 'www.example.com'))
        parsed = parse_url('www.example.com', default_scheme='https')
        self.assertEqual(parsed, URL('https', 'www.example.com'))
        parsed = parse_url('https://www.example.com', default_scheme='https')
        self.assertEqual(parsed, URL('https', 'www.example.com'))

    def test_addr(self):
        parsed = parse_url('www.example.com')
        self.assertEqual(parsed.addr, ('www.example.com', 80))
        parsed = parse_url('https://www.example.com')
        self.assertEqual(parsed.addr, ('www.example.com', 443))

    def test_ssl(self):
        parsed = parse_url('www.example.com')
        self.assertFalse(parsed.ssl)
        parsed = parse_url('http://www.example.com')
        self.assertFalse(parsed.ssl)
        parsed = parse_url('https://www.example.com')
        self.assertTrue(parsed.ssl)

    def test_target(self):
        parsed = parse_url('www.example.com')
        self.assertEqual(parsed.target, '/')
        parsed = parse_url('www.example.com/foo')
        self.assertEqual(parsed.target, '/foo')
        parsed = parse_url('www.example.com?bar')
        self.assertEqual(parsed.target, '/?bar')
        parsed = parse_url('www.example.com/foo?bar')
        self.assertEqual(parsed.target, '/foo?bar')


class TestGetHeader(UnitTest):

    headers = [('foo', 'fooval'),
               ('bar', 'barval'),
               ('baz', 'bazval')]

    def test_simple(self):
        self.assertEqual(get_header(self.headers, 'foo'), 'fooval')
        self.assertEqual(get_header(self.headers, 'bar'), 'barval')
        self.assertEqual(get_header(self.headers, 'baz'), 'bazval')

    def test_case_insensitive(self):
        self.assertEqual(get_header(self.headers, 'Foo'), 'fooval')
        self.assertEqual(get_header(self.headers, 'FOO'), 'fooval')

    def test_not_present(self):
        self.assertIsNone(get_header(self.headers, 'qux'))

    def test_default_value(self):
        self.assertEqual(get_header(self.headers, 'qux', 'quxval'), 'quxval')


class TestRemoveHeaders(UnitTest):

    headers = [('foo', 'fooval1'),
               ('bar', 'barval1'),
               ('foo', 'fooval2'),
               ('baz', 'bazval'),
               ('bar', 'barval2')]

    def test_simple(self):
        self.assertEqual(remove_headers(self.headers[:], 'foo'),
                         [('bar', 'barval1'), ('baz', 'bazval'), ('bar', 'barval2')])
        self.assertEqual(remove_headers(self.headers[:], 'bar'),
                         [('foo', 'fooval1'), ('foo', 'fooval2'), ('baz', 'bazval')])

    def test_in_place(self):
        headers = self.headers[:]
        removed = remove_headers(headers, 'foo')
        self.assertIs(headers, removed)

    def test_non_quadratic(self):
        # Ensure remove_headers() doesn't take quadratic time.
        names = ('foo', 'bar', 'baz', 'qux')
        headers = []
        for i in range(100000):
            name = names[i%4]
            headers.append((name, name + 'val'))
        removed = remove_headers(headers, 'foo')
        self.assertEqual(len(removed), 75000)


class TestHttpProtocol(UnitTest):

    def setUp(self):
        super(TestHttpProtocol, self).setUp()
        self.requests = Queue()

    def store_request(self, message, transport, protocol):
        self.requests.put(message)
        protocol.writer.write(b'HTTP/1.1 200 OK\r\n\r\n')

    def parse_request(self, *chunks):
        # Parse the HTTP request made up of *chunks.
        transport = MockTransport()
        protocol = HttpProtocol(self.store_request, server_side=True)
        transport.start(protocol)
        for chunk in chunks:
            protocol.data_received(chunk)
        self.assertIsNone(protocol._error)
        self.transport = transport
        self.protocol = protocol

    def get_request(self):
        # Get a parsed request.
        m = self.requests.get(timeout=1.0)
        self.assertIsInstance(m, HttpMessage)
        self.assertEqual(m.message_type, http.REQUEST)
        return m

    # Tests that parse a request

    def test_simple_request(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n\r\n'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.version, '1.1')
        self.assertIsNone(m.status_code)
        self.assertEqual(m.method, 'GET')
        self.assertEqual(m.url, '/')
        self.assertTrue(m._should_keep_alive)
        self.assertEqual(m.parsed_url, URL(path='/'))
        self.assertEqual(m.headers, [('Host', 'example.com')])
        self.assertIsInstance(m.body, Stream)
        self.assertTrue(m.body.buffer.eof)

    def test_request_with_body(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Content-Length: 3\r\n\r\nFoo'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.url, '/')
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Host', 'example.com'), ('Content-Length', '3')])
        self.assertFalse(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'Foo')
        self.assertTrue(m.body.buffer.eof)

    def test_request_with_body_incremental(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Content-Length: 3\r\n\r\nFoo'
        self.parse_request(*[r[i:i+1] for i in range(len(r))])
        m = self.get_request()
        self.assertEqual(m.url, '/')
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Host', 'example.com'), ('Content-Length', '3')])
        self.assertFalse(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'Foo')
        self.assertTrue(m.body.buffer.eof)

    def test_request_with_chunked_body(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Transfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\n\r\n'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.url, '/')
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Host', 'example.com'),
                                     ('Transfer-Encoding', 'chunked')])
        self.assertFalse(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'Foo')
        self.assertTrue(m.body.buffer.eof)

    def test_request_with_chunked_body_incremental(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Transfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\n\r\n'
        self.parse_request(*[r[i:i+1] for i in range(len(r))])
        m = self.get_request()
        self.assertEqual(m.url, '/')
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Host', 'example.com'),
                                     ('Transfer-Encoding', 'chunked')])
        self.assertFalse(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'Foo')
        self.assertTrue(m.body.buffer.eof)

    def test_request_with_chunked_body_and_trailers(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Transfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\nETag: foo\r\n\r\n'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.url, '/')
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Host', 'example.com'),
                                     ('Transfer-Encoding', 'chunked'),
                                     ('ETag', 'foo')])
        self.assertFalse(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'Foo')
        self.assertTrue(m.body.buffer.eof)

    def test_pipelined_requests(self):
        r = b'GET /0 HTTP/1.1\r\nHost: example0.com\r\n\r\n' \
            b'GET /1 HTTP/1.1\r\nHost: example1.com\r\n\r\n'
        self.parse_request(r)
        for i in range(2):
            m = self.get_request()
            self.assertEqual(m.url, '/{0}'.format(i))
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.headers, [('Host', 'example{}.com'.format(i))])
            self.assertTrue(m.body.buffer.eof)

    def test_pipelined_requests_with_body(self):
        r = b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Content-Length: 4\r\n\r\nFoo0' \
            b'GET / HTTP/1.1\r\nHost: example.com\r\n' \
            b'Content-Length: 4\r\n\r\nFoo1'
        self.parse_request(r)
        for i in range(2):
            m = self.get_request()
            self.assertEqual(m.url, '/')
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.headers, [('Host', 'example.com'), ('Content-Length', '4')])
            self.assertFalse(m.body.buffer.eof)
            self.assertEqual(m.body.read(), 'Foo{}'.format(i).encode('ascii'))
            self.assertTrue(m.body.buffer.eof)

    def test_request_url(self):
        r = b'GET /foo/bar HTTP/1.1\r\n' \
            b'Host: example.com\r\n\r\n'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.parsed_url, URL(path='/foo/bar'))

    def test_long_request_url(self):
        r = b'GET http://user:pass@example.com:80/foo/bar?baz=qux#quux HTTP/1.1\r\n' \
            b'Host: example.com\r\n\r\n'
        self.parse_request(r)
        m = self.get_request()
        self.assertEqual(m.parsed_url, URL('http', 'example.com', '/foo/bar', 'baz=qux', 'quux',
                                           port='80', userinfo='user:pass'))

    # Tests that parse a response

    def parse_response(self, *chunks, **kwargs):
        # Parse the HTTP resposne made up of *chunks.
        transport = MockTransport()
        protocol = HttpProtocol()
        transport.start(protocol)
        methods = kwargs.get('methods', [])
        if methods:
            protocol._requests = methods
        for chunk in chunks:
            protocol.data_received(chunk)
        self.assertIsNone(protocol._error)
        self.transport = transport
        self.protocol = protocol

    def get_response(self):
        # Get a parsed resposne.
        m = self.protocol.getresponse()
        self.assertIsInstance(m, HttpMessage)
        self.assertEqual(m.message_type, http.RESPONSE)
        return m

    def test_simple_response(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Content-Length', '0')])
        self.assertEqual(m.get_header('Content-Length'), '0')
        self.assertEqual(m.body.read(), b'')
        self.assertTrue(m.body.buffer.eof)

    def test_response_with_body(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 3\r\n\r\nFoo'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Content-Length', '3')])
        self.assertEqual(m.get_header('Content-Length'), '3')
        self.assertEqual(m.body.read(), b'Foo')
        self.assertEqual(m.body.read(), b'')

    def test_response_with_body_incremental(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 3\r\n\r\nFoo'
        self.parse_response([r[i:i+1] for i in range(len(r))])
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Content-Length', '3')])
        self.assertEqual(m.body.read(), b'Foo')
        self.assertEqual(m.body.read(), b'')

    def test_response_with_chunked_body(self):
        r = b'HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\n\r\n'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Transfer-Encoding', 'chunked')])
        self.assertEqual(m.body.read(), b'Foo')
        self.assertEqual(m.body.read(), b'')

    def test_response_with_chunked_body_incremental(self):
        r = b'HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\n\r\n'
        self.parse_response([r[i:i+1] for i in range(len(r))])
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Transfer-Encoding', 'chunked')])
        self.assertEqual(m.body.read(), b'Foo')
        self.assertEqual(m.body.read(), b'')

    def test_response_with_chunked_body_and_trailers(self):
        r = b'HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n' \
            b'3\r\nFoo\r\n0\r\nETag: foo\r\n\r\n'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.status_code, 200)
        self.assertEqual(m.headers, [('Transfer-Encoding', 'chunked'), ('ETag', 'foo')])
        self.assertEqual(m.body.read(), b'Foo')
        self.assertEqual(m.body.read(), b'')

    def test_pipelined_responses(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 0\r\nSet-Cookie: foo=0\r\n\r\n' \
            b'HTTP/1.1 200 OK\r\nContent-Length: 0\r\nSet-Cookie: foo=1\r\n\r\n'
        self.parse_response(r)
        for i in range(2):
            m = self.get_response()
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.status_code, 200)
            cookie = 'foo={0}'.format(i)
            self.assertEqual(m.headers, [('Content-Length', '0'), ('Set-Cookie', cookie)])
            self.assertEqual(m.body.read(), b'')

    def test_pipelined_responses_with_body(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nFoo0' \
            b'HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nFoo1'
        self.parse_response(r)
        for i in range(2):
            m = self.get_response()
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.status_code, 200)
            self.assertEqual(m.headers, [('Content-Length', '4')])
            self.assertEqual(m.body.read(), 'Foo{0}'.format(i).encode('ascii'))
            self.assertEqual(m.body.read(), b'')

    def test_pipelined_head_responses(self):
        r = b'HTTP/1.1 200 OK\r\nContent-Length: 3\r\n\r\n' \
            b'HTTP/1.1 200 OK\r\nContent-Length: 3\r\n\r\n'
        self.parse_response(r, methods=['HEAD', 'HEAD'])
        for i in range(2):
            m = self.get_response()
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.status_code, 200)
            self.assertEqual(m.headers, [('Content-Length', '3')])
            self.assertEqual(m.body.read(), b'')

    def test_pipelined_empty_responses(self):
        # These status codes take no body. The parser should know this.
        r = b'HTTP/1.1 100 OK\r\nCookie: foo0\r\n\r\n' \
            b'HTTP/1.1 204 OK\r\nCookie: foo1\r\n\r\n' \
            b'HTTP/1.1 304 OK\r\nCookie: foo2\r\n\r\n'
        self.parse_response(r)
        for i in range(3):
            m = self.get_response()
            self.assertEqual(m.version, '1.1')
            self.assertEqual(m.headers, [('Cookie', 'foo{0}'.format(i))])
            self.assertEqual(m.body.read(), b'')

    def test_pipelined_200_response_eof(self):
        # No content-length so 200 requires and EOF to indicate the end of
        # message. The second request is therefore interpreted as a the body of
        # the first.
        r = b'HTTP/1.1 200 OK\r\nCookie: foo0\r\n\r\nfoo'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Cookie', 'foo0')])
        self.assertEqual(m.body.read(3), b'foo')
        self.assertFalse(m.body.buffer.eof)
        self.protocol.connection_lost(None)
        self.assertTrue(m.body.buffer.eof)
        self.assertEqual(m.body.read(), b'')

    def test_pipelined_204_response_eof(self):
        # A 204 never has a body so the absence of a Content-Length header
        # still does not require it to see an EOF.
        r = b'HTTP/1.1 204 OK\r\nCookie: foo0\r\n\r\n'
        self.parse_response(r)
        m = self.get_response()
        self.assertEqual(m.version, '1.1')
        self.assertEqual(m.headers, [('Cookie', 'foo0')])
        self.assertEqual(m.body.read(), b'')


def hello_app(environ, start_response):
    headers = [('Content-Type', 'text/plain')]
    start_response('200 OK', headers)
    return [b'Hello!']


def echo_app(environ, start_response):
    headers = [('Content-Type', 'text/plain')]
    for name in environ:
        if name.startswith('HTTP_X'):
            hname = name[5:].replace('_', '-').title()
            headers.append((hname, environ[name]))
    headers.sort()  # for easier comparison
    body = environ['wsgi.input'].read()
    start_response('200 OK', headers)
    return [body]


class TestHttp(UnitTest):

    def test_simple(self):
        server = HttpServer(hello_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        client.request('GET', '/')
        resp = client.getresponse()
        self.assertEqual(resp.version, '1.1')
        self.assertEqual(resp.status_code, 200)
        serv = resp.get_header('Server')
        self.assertTrue(serv.startswith('gruvi'))
        ctype = resp.get_header('Content-Type')
        self.assertEqual(ctype, 'text/plain')
        self.assertEqual(resp.body.read(), b'Hello!')
        server.close()
        client.close()

    def test_simple_pipe(self):
        server = HttpServer(hello_app)
        server.listen(self.pipename())
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        client.request('GET', '/')
        resp = client.getresponse()
        self.assertEqual(resp.body.read(), b'Hello!')
        server.close()
        client.close()

    def test_simple_ssl(self):
        server = HttpServer(hello_app)
        context = self.get_ssl_context()
        server.listen(('localhost', 0), ssl=context)
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr, ssl=context)
        client.request('GET', '/')
        resp = client.getresponse()
        self.assertEqual(resp.body.read(), b'Hello!')
        server.close()
        client.close()

    def test_request_headers(self):
        server = HttpServer(echo_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        client.request('GET', '/', headers=[('X-Echo', 'Bar')])
        resp = client.getresponse()
        self.assertEqual(resp.get_header('X-Echo'), 'Bar')
        body = resp.body.read()
        self.assertEqual(body, b'')
        server.close()
        client.close()

    def test_request_body(self):
        server = HttpServer(echo_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        client.request('POST', '/', body=b'foo')
        resp = client.getresponse()
        body = resp.body.read()
        self.assertEqual(body, b'foo')
        server.close()
        client.close()

    def test_request_body_sequence(self):
        server = HttpServer(echo_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        client.request('POST', '/', body=[b'foo', b'bar'])
        resp = client.getresponse()
        body = resp.body.read()
        self.assertEqual(body, b'foobar')
        server.close()
        client.close()

    def test_request_trailers(self):
        server = HttpServer(echo_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = HttpClient()
        client.connect(addr)
        headers = [('X-Echo', 'Bar'), ('Trailer', 'X-Length')]
        def genbody():
            yield b'FooBody'
            headers.append(('X-Length', '7'))
        client.request('GET', '/', body=genbody(), headers=headers)
        resp = client.getresponse()
        self.assertEqual(resp.get_header('X-Echo'), 'Bar')
        self.assertEqual(resp.get_header('X-Length'), '7')
        body = resp.body.read()
        self.assertEqual(body, b'FooBody')
        server.close()
        client.close()

    def test_illegal_request(self):
        server = HttpServer(hello_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = StreamClient()
        client.connect(addr)
        client.write(b'foo\r\n')
        client.write_eof()
        buf = client.read()
        self.assertEqual(buf, b'')
        server.close()
        client.close()


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

"""
The :mod:`gruvi.jsonrpc` module implements a JSON-RPC client and server.

There are two version of JSON-RPC: version 1.0 and version 2.0. While these
versions are similar, they are not mutually compatible. The classes exposed by
this module implement both versions. The default version to use is stored in
:attr:`JsonRpcProtocol.default_version`, and most constructors take an argument
that can be used to override this.

The "batch" feature of version 2.0 is not supported. It more relevant for
JSON-RPC over HTTP rather for that clients and servers that operate directly on
top of a connection.

The two main classes in this module are :class:`JsonRpcClient` and
:class:`JsonRpcServer`. The difference is only who initiates the connection at
the transport level. The JSON-RPC protocol itself does not distinguish between
clients and servers.

Both the client and the server can receive and respond to incoming messages.
These may be method calls (more common for servers), or notifications (common
for both client and servers). These incoming messages are handled by a message
handler, which is mandatory for a server and optional for a client. Note that
for simple blocking method callls from client to a server no message handler is
needed.

Message handlers run in a separate dispatcher fiber, one per connection. This
means that a client will have at most one dispatcher fiber, while a server will
have exactly one fiber per connection. Because they run in a separate fiber,
message handlers can call into switchpoints themselves.
"""

from __future__ import absolute_import, print_function

import json
import functools
import six

from . import compat
from .hub import switchpoint, switch_back
from .util import delegate_method, Absent
from .transports import TransportError
from .protocols import ProtocolError, MessageProtocol
from .stream import Stream
from .endpoints import Client, Server
from .jsonrpc_ffi import lib, ffi

__all__ = ['JsonRpcError', 'JsonRpcVersion', 'JsonRpcProtocol',
           'JsonRpcClient', 'JsonRpcServer']

# Message types

REQUEST = 1   #: Constant indicating a JSON-RPC request message.
RESPONSE = 2  #: Constant indicating a JSON-RPC response message.

# JSON-RPC error codes. In both versions the "error" field is required to be an
# Object, but v2 specifies a structure for it. We use the v2 error format for
# both versions.

errorcode = {}  #: Dictionary mapping error codes to error names.
_errlist = {}
_errobjs = {}

def add_error(code, name, message):
    globals()[name] = code
    errorcode[code] = name
    _errlist[code] = message
    _errobjs[code] = {'code': code, 'message': message}

add_error(-32000, 'SERVER_ERROR', 'Server error')
add_error(-32001, 'MESSAGE_TOO_LARGE', 'Message too large')
add_error(-32002, 'UNKNOWN_RESPONSE', 'Unknown response')
add_error(-32600, 'INVALID_REQUEST', 'Invalid request')
add_error(-32601, 'METHOD_NOT_FOUND', 'Method not found')
add_error(-32602, 'INVALID_PARAMS', 'Invalid parameters')
add_error(-32603, 'INTERNAL_ERROR', 'Internal error')
add_error(-32700, 'PARSE_ERROR', 'Parse error')

del add_error

def strerror(code):
    """Return an error message for error *code*.

    The error code is the ``error.code`` field of a JSON-RPC response message.
    The JSON-RPC version 2 spec contains a list of standard error codes.
    """
    return _errlist.get(code, 'No error description available')


class JsonRpcError(ProtocolError):
    """Exception that is raised in case of JSON-RPC errors."""

    @property
    def error(self):
        """The ``error`` field from the JSON-RPC response, if available."""
        return self.args[1] if len(self.args) > 0 else None


# A class that encapsulates the differences between the two JSON-RPC versions
# that exist out there (1.0 and 2.0).

class JsonRpcVersion(object):
    """Class that encapsulates the differences between JSON-RPC vesion 1.0 and 2.0."""

    _id_template = 'gruvi.{}'

    def __init__(self, name):
        self._next_id = 0
        self._name = name

    @property
    def name(self):
        """Return the JSON-RPC version as a string, e.g. `'1.0'`."""
        return self._name

    def next_id(self):
        """Return a unique message ID."""
        msgid = self._id_template.format(self._next_id)
        self._next_id += 1
        return msgid

    @staticmethod
    def create(version):
        """Return a new instance for *version*, which can be either `'1.0'`
        or `'2.0'`."""
        clsname = 'JsonRpcV{}'.format(version.rstrip('.0'))
        cls = globals()[clsname]
        return cls(version)

    def check_message(self, message):
        """Check *message* and return the message type. Raise a `ValueError` in
        case of an invalid message."""
        raise NotImplementedError

    def create_request(self, method, args=None, notification=False):
        """Create a new request message for *method* with arguments *args*.

        The optional *notification* argument, if nonzero, creates a
        notification and no response to this message is expected.
        """
        raise NotImplementedError

    def create_response(self, request, result=None, error=None):
        """Create a new response message, as a response to *request*.

        A successful response is created when the optional *error* argument is
        not provided. In this case *result* specifies the result, which may be
        any value including ``None``. If *error* is provided, an error response
        is created, and *error* should be a dict as specified by the JSON-RPC
        specifications.

        The *request* argument may be explicitly set to ``None``, in which case
        this is an error response that is not specific to any one request.
        """
        raise NotImplementedError


class JsonRpcV1(JsonRpcVersion):
    """JSON-RPC version 1 message routines.

    See http://json-rpc.org/wiki/specification.
    """

    def check_message(self, message):
        msgid = message.get('id', Absent)
        if msgid is Absent:
            raise ValueError('"id" is absent')
        method = message.get('method', Absent)
        # If "method" is absent, it's a response. Otherwise it's a request.
        if method is Absent:
            result = message.get('result', Absent)
            error = message.get('error', Absent)
            if result is Absent or error is Absent or \
                    result is not None and error is not None or \
                    error is not None and not isinstance(error, dict) or \
                    len(message) != 3:
                raise ValueError('illegal "result" or "error" field')
            return RESPONSE
        elif isinstance(method, six.string_types):
            params = message.get('params', Absent)
            if not isinstance(params, (tuple, list)) or len(message) != 3:
                raise ValueError('illegal "param" field')
            return REQUEST
        else:
            raise ValueError('illegal "method" field')

    def create_request(self, method, args=None, notification=False):
        message = {'id': self.next_id() if not notification else None,
                   'method': method,
                   'params': [] if args is None else args}
        return message

    def create_response(self, request, result=None, error=None):
        message = {'id': request['id'] if request else None,
                   'result': result,
                   'error': error}
        return message


class JsonRpcV2(JsonRpcVersion):
    """JSON-RPC version 2 message routines.

    See http://www.jsonrpc.org/specification.
    """

    def check_message(self, message):
        version = message.get('jsonrpc')
        if version != '2.0':
            raise ValueError('illegal "version" field')
        msgid = message.get('id', Absent)
        if msgid is not Absent and msgid is not None and \
                    not isinstance(msgid, six.string_types) and \
                    not isinstance(msgid, (int, float)):
            raise ValueError('illegal "id" field')
        # If "method" is absent, it's a response. Otherwise it's a request.
        method = message.get('method', Absent)
        if method is Absent:
            result = message.get('result', Absent)
            error = message.get('error', Absent)
            if msgid is Absent or \
                    error is not Absent and not isinstance(error, dict) or \
                    (result is Absent) == (error is Absent) or \
                    len(message) != 3:
                raise ValueError('illegal "result" or "error" field')
            return RESPONSE
        elif isinstance(method, six.string_types):
            params = message.get('params', Absent)
            if params is not Absent and not isinstance(params, (dict, tuple, list)) or \
                    len(message) != 2 + (msgid is not Absent) + (params is not Absent):
                raise ValueError('illegal "param" field')
            return REQUEST
        else:
            raise ValueError('illegal "method" field')

    def create_request(self, method, args=None, notification=False):
        message = {'jsonrpc': '2.0',
                   'method': method,
                   'params': [] if args is None else args}
        if not notification:
            message['id'] = self.next_id()
        return message

    def create_response(self, request, result=None, error=None):
        message = {'jsonrpc': '2.0',
                   'id': request['id'] if request else None}
        if error is None:
            message['result'] = result
        else:
            message['error'] = error
        return message


def message_info(message):
    """Return a string describing a message, for debugging purposes."""
    method = message.get('method')
    msgid = message.get('id')
    error = message.get('error')
    if method and msgid is not None:
        return 'method call "{}", id = "{}"'.format(method, msgid)
    elif method:
        return 'notification "{}"'.format(method)
    elif error is not None and msgid is not None:
        return 'error reply to id = "{}"'.format(msgid)
    elif error is not None:
        code = error.get('code', '(none)')
        return 'error reply: {}'.format(errorcode.get(code, code))
    else:
        return 'method return for id = "{}"'.format(msgid)


def serialize(msg):
    """Serialize a JSON-RPC message."""
    return json.dumps(msg, indent=2).encode('utf-8')


# JSON-RPC protocol implementation.

class JsonRpcProtocol(MessageProtocol):
    """A JSON-RPC client and server :class:`~gruvi.Protocol` implementation."""

    #: Class level attribute that specifies the default timeout.
    default_timeout = 30

    #: Class level attribute that specifies the default JSON-RPC version.
    default_version = '2.0'

    # Any message larger than this and the connection is dropped.
    # This can be increased, but do note that message are kept in memory until
    # they are dispatched.
    max_message_size = 65536

    def __init__(self, handler=None, version=None, timeout=None):
        super(JsonRpcProtocol, self).__init__(handler, timeout=timeout)
        self._handler = handler
        self._version = self.default_version if version is None else version
        self._timeout = self.default_timeout if timeout is None else timeout
        self._buffer = bytearray()
        self._context = ffi.new('struct split_context *')
        self._method_calls = {}
        self._version = JsonRpcVersion.create(self._version)

    @property
    def version(self):
        """Return the :class:`JsonRpcVersion` instance for this protocol."""
        return self._version

    def connection_made(self, transport):
        # Protocol callback
        super(JsonRpcProtocol, self).connection_made(transport)
        self._writer = Stream(transport, 'w')

    def connection_lost(self, exc):
        # Protocol callback
        super(JsonRpcProtocol, self).connection_lost(exc)
        for switcher in self._method_calls.values():
            switcher.throw(TransportError('connection lost'))
        self._method_calls.clear()

    def data_received(self, data):
        # Protocol callback
        if self._error:
            self._log.debug('ignore data received after error')
            return
        offset = 0
        error = None
        # Set up the CFFI data structure to communicate with the splitter.
        # The "cdata" local variable keeps the data alive during this function.
        cdata = self._context.buf = ffi.from_buffer(data)
        self._context.buflen = len(data)
        self._context.offset = 0
        # Use the CFFI JSON splitter to delineate JSON messages in the input
        # stream. Then decode, parse, check and route them.
        while offset != len(data):
            split_error = lib.json_split(self._context)
            if split_error and split_error != lib.INCOMPLETE:
                self._log.debug('message split error: {}', split_error)
                error = PARSE_ERROR
                break
            # We have a good (partial) message. Check the size, including our buffer.
            size = len(self._buffer) + self._context.offset - offset
            if size > self.max_message_size:
                self._log.debug('message too large: {} bytes', size)
                error = MESSAGE_TOO_LARGE
                break
            # If we don't have a full message, save what we have and return.
            if split_error == lib.INCOMPLETE:
                assert self._context.offset == len(data)
                self._buffer.extend(data[offset:])
                break
            # A full message has been received. Process it.
            chunk = data[offset:self._context.offset]
            if self._buffer:
                # Prepend partial message from last parser invocation.
                self._buffer.extend(chunk)
                chunk = self._buffer
                self._buffer = bytearray()
            # Now parse it. Assume UTF-8 encoding. While the JSON spec allows
            # for UTF-16 and UTF-32, it makes it clear that those are not
            # interoperable (RFC7159, section 8.1).
            try:
                decoded = chunk.decode('utf8')
                message = json.loads(decoded)
            except Exception as e:
                self._log.debug('message parse error: {}', e)
                error = PARSE_ERROR
                break
            try:
                mtype = self._version.check_message(message)
            except ValueError as e:
                self._log.debug('illegal JSON-RPC message: {}', e)
                error = INVALID_REQUEST
                break
            self._log.debug(message_info(message))
            self._log.trace('\n\n{}\n', chunk)
            # Now route the message to its correct destination.
            if mtype == RESPONSE and message.get('id') in self._method_calls:
                # Response to a method call issued via call_method(): wake up caller
                switcher = self._method_calls.pop(message['id'])
                switcher(message)
            elif mtype == REQUEST and message.get('method') and not self._handler:
                # Method call but there is no handler: return error
                message = self._version.create_response(message, None, _errobjs[NO_SUCH_METHOD])
                self._transport.write(serialize(message))
            elif self._handler:
                # Everything else goes to the handler, if there is one.
                self._queue.put_nowait(message)
                self._maybe_pause_transport()
            offset = self._context.offset
        if error:
            message = self._version.create_response(None, None, _errobjs[error])
            self._transport.write(serialize(message))
            self._transport.close()
            self._error = JsonRpcError('Transport closed after protocol violation')

    @switchpoint
    def send_message(self, message):
        """Send a raw JSON-RPC message.

        The *message* argument must be a dictionary containing a valid JSON-RPC
        message according to the version passed into the constructor.
        """
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._transport is None:
            raise JsonRpcError('not connected')
        self._version.check_message(message)
        self._writer.write(serialize(message))

    @switchpoint
    def call_method(self, method, *args):
        """Call a JSON-RPC method and wait for its result.

        The *method* is called with positional arguments *args*.

        On success, the ``result`` field from the JSON-RPC response is
        returned.  On error, a :class:`JsonRpcError` is raised, which you can
        use to access the ``error`` field of the JSON-RPC response.
        """
        message = self._version.create_request(method, args)
        msgid = message['id']
        try:
            with switch_back(self._timeout) as switcher:
                self._method_calls[msgid] = switcher
                self.send_message(message)
                args, _ = self._hub.switch()
        finally:
            self._method_calls.pop(msgid, None)
        response = args[0]
        assert response['id'] == msgid
        error = response.get('error')
        if error is not None:
            raise JsonRpcError('error response calling "{}"'.format(method), error)
        return response['result']

    @switchpoint
    def send_notification(self, method, *args):
        """Send a JSON-RPC notification.

        The notification *method* is sent with positional arguments *args*.
        """
        message = self._version.create_request(method, args, notification=True)
        self.send_message(message)

    @switchpoint
    def send_response(self, request, result=None, error=None):
        """Respond to a JSON-RPC method call.

        This is a response to the message in *request*. If *error* is not
        provided, then this is a succesful response, and the value in *result*,
        which may be ``None``, is passed back to the client. if *error* is
        provided and not ``None`` then an error is sent back. In this case
        *error* must be a dictionary as specified by the JSON-RPC spec.
        """
        message = self._version.create_response(request, result, error)
        self.send_message(message)


class JsonRpcClient(Client):
    """A JSON-RPC :class:`~gruvi.Client`."""

    def __init__(self, handler=None, version=None, timeout=None):
        """
        The *handler* argument specifies an optional JSON-RPC message handler.
        You need to supply a message handler if you want to listen to
        notifications or you want to respond to server-to-client method calls.
        If provided, the message handler it must be a callable with signature
        ``handler(message, transport, protocol)``.

        The *version* and *timeout* argument can be used to override the
        default protocol version and timeout, respectively.
        """
        protocol_factory = functools.partial(JsonRpcProtocol, handler, version)
        super(JsonRpcClient, self).__init__(protocol_factory, timeout)

    protocol = Client.protocol

    delegate_method(protocol, JsonRpcProtocol.send_message)
    delegate_method(protocol, JsonRpcProtocol.call_method)
    delegate_method(protocol, JsonRpcProtocol.send_notification)


class JsonRpcServer(Server):
    """A JSON-RPC :class:`~gruvi.Server`."""

    max_connections = 1000

    def __init__(self, handler, version=None, timeout=None):
        """
        The *handler* argument specifies the JSON-RPC message handler. It must
        be a callable with signature ``handler(message, transport, protocol)``.
        The message handler is called in a separate dispatcher fiber (one per
        connection).

        The *version* and *timeout* argument can be used to override the
        default protocol version and timeout, respectively.
        """
        protocol_factory = functools.partial(JsonRpcProtocol, handler, version)
        super(JsonRpcServer, self).__init__(protocol_factory, timeout=timeout)



# Gruvi example program: a cURL like URL downloader

import sys
import argparse
from gruvi.http import HttpClient, parse_url

parser = argparse.ArgumentParser()
parser.add_argument('url')
args = parser.parse_args()

url = parse_url(args.url)

client = HttpClient()
client.connect(url.addr, ssl=url.ssl)
client.request('GET', url.target)

resp = client.getresponse()
if not 200 <= resp.status_code <= 399:
    sys.stderr.write('Error: got status {}\n'.format(resp.status_code))
    sys.exit(1)

charset = resp.charset or 'iso8859-1'

while True:
    buf = resp.body.read(4096)
    if not buf:
        break
    sys.stdout.write(buf.decode(charset))



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import fibers
import threading
import heapq

from .hub import switchpoint, get_hub, switch_back, assert_no_switchpoints
from .callbacks import add_callback, remove_callback, pop_callback
from .callbacks import run_callbacks, walk_callbacks

__all__ = ['Lock', 'RLock', 'Event', 'Condition', 'QueueEmpty', 'QueueFull',
           'Queue', 'LifoQueue', 'PriorityQueue']


# All primitives in this module a thread safe!


class _Lock(object):
    """Base class for regular and reentrant locks."""

    __slots__ = ('_reentrant', '_lock', '_locked', '_owner', '_callbacks')

    def __init__(self, reentrant):
        # Allocate a new lock
        self._reentrant = reentrant
        self._lock = threading.Lock()
        self._locked = 0
        self._owner = None
        self._callbacks = None

    def locked(self):
        """Whether the lock is currently locked."""
        return self._locked

    @switchpoint
    def acquire(self, blocking=True, timeout=None):
        """Acquire the lock.

        If *blocking* is true (the default), then this will block until the
        lock can be acquired. The *timeout* parameter specifies an optional
        timeout in seconds.

        The return value is a boolean indicating whether the lock was acquired.
        """
        hub = get_hub()
        try:
            # switcher.__call__ needs to be synchronized with a lock IF it can
            # be called from different threads. This is the case here because
            # this method may be called from multiple threads and the callbacks
            # are run in the calling thread. So pass it our _lock.
            with switch_back(timeout, lock=self._lock) as switcher:
                with self._lock:
                    if not self._locked:
                        self._locked = 1
                        self._owner = fibers.current()
                        return True
                    elif self._reentrant and self._owner is fibers.current():
                        self._locked += 1
                        return True
                    elif not blocking:
                        return False
                    handle = add_callback(self, switcher)
                # It is safe to call hub.switch() outside the lock. Another
                # thread could have called acquire()+release(), thereby firing
                # the switchback. However the switchback only schedules the
                # switchback in our hub, it won't execute it yet. So the
                # switchback won't actually happen until we switch to the hub.
                hub.switch()
                # Here the lock should be ours because _release() wakes up only
                # the fiber that it passed the lock.
                assert self._locked > 0
                assert self._owner is fibers.current()
        except BaseException as e:
            # Likely a Timeout but could also be e.g. Cancelled
            with self._lock:
                # Clean up the callback. It might have been popped by
                # _release() but that is OK.
                remove_callback(self, handle)
                # This fiber was passed the lock but before that an exception
                # was already scheduled with run_callback() (likely through
                # Fiber.throw())
                if self._owner is fibers.current():
                    self._release()
            if e is switcher.timeout:
                return False
            raise
        return True

    def _release(self):
        # Low-level release. Lock must be held.
        if self._locked > 1:
            self._locked -= 1
            return
        while self._callbacks:
            switcher, _ = pop_callback(self)
            if switcher.active:
                self._owner = switcher.fiber
                switcher.switch()
                return
        self._owner = None
        self._locked = 0

    def release(self):
        """Release the lock."""
        with self._lock:
            if not self._locked:
                raise RuntimeError('lock not currently held')
            elif self._reentrant and self._owner is not fibers.current():
                raise RuntimeError('lock not owned by this fiber')
            self._release()

    __enter__ = acquire
    __exit__ = lambda self, *exc_info: self.release()

    # Internal API used by Condition instances.

    def _acquire_restore(self, state):
        # Acquire a lock and restore the owner and lock count.
        self.acquire()
        self._owner, self._locked = state

    def _release_save(self):
        # Release a lock even if it is locked multiple times. Return the state.
        state = self._owner, self._locked
        self.release()
        return state


class Lock(_Lock):
    """A lock.

    The lock can be locked and unlocked explicitly using :meth:`acquire` and
    :meth:`release`, and it can also be used as a context manager.
    """

    __slots__ = _Lock.__slots__

    def __init__(self):
        super(Lock, self).__init__(False)


class RLock(_Lock):
    """A reentrant lock.

    A reentrant lock has the notion of a "lock owner" and a "lock count". If a
    reentrant lock is acquired, and it was already acquired by the current
    fiber, then the lock count is increased and the acquire call will be
    successful. Unlocking a reentrant lock may only be done by the lock owner.
    The lock becomes unlocked only after it is released as many times as it was
    acquired.
    """

    __slots__ = _Lock.__slots__

    def __init__(self):
        super(RLock, self).__init__(True)


# A few words on the use of fiber locks (Lock) vs thread locks (threading.Lock)
# in the code below.
#
# There is no difference between both locks from a safety point of view. Both
# locks are thread-safe (which implies they are fiber-safe as well). The
# difference is who gets blocked when trying to acquire a lock that is already
# locked. With a fiber lock only the current fiber is blocked and other fibers
# in current thread can continue (and fibers in other threads as well, of
# course). With a thread lock the entire current thread is blocked including
# all its fibers.
#
# This means that if we never call hub.switch() when a lock is held, fiber and
# thread locks are completely identical. In this case there's a benefit in
# using thread locks because i) they are smaller and faster, and ii) it makes
# it possible for non-switchpoints to acquire the lock. An example of the
# latter case is Queue.put_nowait().


class Event(object):
    """An event.

    An event contains an internal flag that is initially False. The flag can be
    set using the :meth:`set` method and cleared using the :meth:`clear`
    method.  Fibers can wait for the flag to become set using :meth:`wait`.

    Events are level triggered, meaning that the condition set by :meth:`set`
    is "sticky". Setting the event will unblock all current waiters and will
    cause future calls to :meth:`wait` not to block, until :meth:`clear` is
    called again.
    """

    __slots__ = ('_flag', '_lock', '_callbacks')

    def __init__(self):
        self._flag = False
        self._lock = threading.Lock()
        self._callbacks = None

    def is_set(self):
        return self._flag

    def set(self):
        """Set the internal flag, and wake up any fibers blocked on :meth:`wait`."""
        with self._lock:
            if self._flag:
                return
            self._flag = True
            with assert_no_switchpoints():
                run_callbacks(self)

    def clear(self):
        """Clear the internal flag."""
        with self._lock:
            self._flag = False

    @switchpoint
    def wait(self, timeout=None):
        """If the internal flag is set, return immediately. Otherwise block
        until the flag gets set by another fiber calling :meth:`set`."""
        # Optimization for the case the Event is already set.
        if self._flag:
            return True
        hub = get_hub()
        try:
            with switch_back(timeout, lock=self._lock) as switcher:
                with self._lock:
                    # Need to check the flag again, now under the lock.
                    if self._flag:
                        return True
                    # Allow other fibers to wake us up via callback in set().
                    # The callback goes to switcher.switch directly() instead of
                    # __call__(), because the latter would try to lock our lock
                    # which is already held when callbacks are run by set().
                    handle = add_callback(self, switcher.switch)
                # See note in Lock.acquire() why we can call to hub.switch()
                # outside the lock.
                hub.switch()
        except BaseException as e:
            with self._lock:
                remove_callback(self, handle)
            if e is switcher.timeout:
                return False
            raise
        return True

    # Support for wait()

    def add_done_callback(self, callback, *args):
        with self._lock:
            if self._flag:
                callback(*args)
                return
            return add_callback(self, callback, args)

    def remove_done_callback(self, handle):
        with self._lock:
            remove_callback(self, handle)


# Utility functions for a condition to work with both Locks and RLocks.

def is_locked(lock):
    """Return whether a lock is locked.

    Suppors :class:`Lock`, :class:`RLock`, :class:`threading.Lock` and
    :class:`threading.RLock` instances.
    """
    if hasattr(lock, 'locked'):
        return lock.locked()
    elif hasattr(lock, '_is_owned'):
        return lock._is_owned()
    else:
        raise TypeError('expecting Lock/RLock')

def acquire_restore(lock, state):
    """Acquire a lock and restore its state."""
    if hasattr(lock, '_acquire_restore'):
        lock._acquire_restore(state)
    elif hasattr(lock, 'acquire'):
        lock.acquire()
    else:
        raise TypeError('expecting Lock/RLock')

def release_save(lock):
    """Release a lock and return its state."""
    if hasattr(lock, '_release_save'):
        return lock._release_save()
    elif hasattr(lock, 'release'):
        lock.release()
    else:
        raise TypeError('expecting Lock/RLock')

def thread_lock(lock):
    """Return the thread lock for *lock*."""
    if hasattr(lock, '_lock'):
        return lock._lock
    elif hasattr(lock, 'acquire'):
        return lock
    else:
        raise TypeError('expecting Lock/RLock')


class Condition(object):
    """A condition.

    A condition is always associated with a lock. The state of the condition
    may only change when the caller has acquired the lock. While the lock is
    held, a condition can be waited for using :meth:`wait`. The wait method
    will release the lock just before blocking itself, so that another fiber
    can call :meth:`notify` to notify the condition.

    The difference between a condition and an :class:`Event` is that a
    condition is edge-trigerred. This means that when a condition is notified,
    only fibers that are waiting *at the time of notification* are unblocked.
    Any fiber that calls :meth:`wait` after the notification, will block until
    the next notification. This also explains why a lock is needed. Without the
    lock there would be a race condition between notification and waiting.
    """

    __slots__ = ('_lock', '_callbacks')

    def __init__(self, lock=None):
        """
        The *lock* argument can be used to share a lock between multiple
        conditions. It must be a :class:`Lock` or :class:`RLock` instance. If
        no lock is provided, a :class:`RLock` is allocated.
        """
        self._lock = lock or RLock()
        self._callbacks = None

    acquire = lambda self, *args: self._lock.acquire(*args)
    release = lambda self: self._lock.release()
    __enter__ = lambda self: self._lock.acquire()
    __exit__ = lambda self, *exc_info: self.release()

    def notify(self, n=1):
        """Raise the condition and wake up fibers waiting on it.

        The optional *n* parameter specifies how many fibers will be notified.
        By default, one fiber is notified.
        """
        if not is_locked(self._lock):
            raise RuntimeError('lock is not locked')
        notified = [0]  # Work around lack of "nonlocal" in py27
        def walker(switcher, predicate):
            if not switcher.active:
                return False  # not not keep switcher that timed out
            if predicate and not predicate():
                return True
            if n >= 0 and notified[0] >= n:
                return True
            switcher.switch()
            notified[0] += 1
            return False  # only notify once
        walk_callbacks(self, walker)

    def notify_all(self):
        """Raise the condition and wake up all fibers waiting on it."""
        self.notify(-1)

    @switchpoint
    def wait(self, timeout=None):
        """Wait for the condition to be notified.

        The return value is True, unless a timeout occurred in which case the
        return value is False.

        The lock must be held before calling this method. This method will
        release the lock just before blocking itself, and it will re-acquire it
        before returning.
        """
        return self.wait_for(None, timeout)

    @switchpoint
    def wait_for(self, predicate, timeout=None):
        """Like :meth:`wait` but additionally for *predicate* to be true.

        The *predicate* argument must be a callable that takes no arguments.
        Its result is interpreted as a boolean value.
        """
        if not is_locked(self._lock):
            raise RuntimeError('lock is not locked')
        hub = get_hub()
        try:
            with switch_back(timeout, lock=thread_lock(self._lock)) as switcher:
                handle = add_callback(self, switcher, predicate)
                # See the comment in Lock.acquire() why it is OK to release the
                # lock here before calling hub.switch().
                # Also if this is a reentrant lock make sure it is fully released.
                state = release_save(self._lock)
                hub.switch()
        except BaseException as e:
            with self._lock:
                remove_callback(self, handle)
            if e is switcher.timeout:
                return False
            raise
        finally:
            acquire_restore(self._lock, state)
        return True


class QueueEmpty(Exception):
    """Queue is empty."""

class QueueFull(Exception):
    """Queue is full."""


class Queue(object):
    """A synchronized FIFO queue. """

    __slots__ = ('_maxsize', '_unfinished_tasks', '_heap', '_size', '_counter',
                 '_lock', '_notempty', '_notfull', '_alldone')

    def __init__(self, maxsize=0):
        """
        The *maxsize* argument specifies the maximum queue size. If it is less
        than or equal to zero, the queue size is infinite.
        """
        self._maxsize = maxsize
        self._unfinished_tasks = 0
        # Use a list/heapq even for a FIFO instead of a deque() because of the
        # latter's high memory use (see comment in Lock). For most protocols
        # there will be one Queue per connection so a low memory footprint is
        # very important.
        self._heap = []
        self._size = 0
        self._counter = 0
        # Use a threading.Lock so that put_nowait() and get_nowait() don't need
        # to be a switchpoint. Also it is more efficient.
        self._lock = threading.Lock()
        self._notempty = Condition(self._lock)
        self._notfull = Condition(self._lock)
        self._alldone = Condition(self._lock)

    def _get_item_priority(self, item):
        # Priority function: FIFO queue by default
        self._counter += 1
        return self._counter

    def qsize(self):
        """Return the size of the queue, which is the sum of the size of all
        its elements."""
        return self._size

    empty = lambda self: self.qsize() == 0
    full = lambda self: self.qsize() >= self.maxsize > 0

    maxsize = property(lambda self: self._maxsize)
    unfinished_tasks = property(lambda self: self._unfinished_tasks)

    @switchpoint
    def put(self, item, block=True, timeout=None, size=None):
        """Put *item* into the queue.

        If the queue is currently full and *block* is True (the default), then
        wait up to *timeout* seconds for space to become available. If no
        timeout is specified, then wait indefinitely.

        If the queue is full and *block* is False or a timeout occurs, then
        raise a :class:`QueueFull` exception.

        The optional *size* argument may be used to specify a custom size for
        the item. The total :meth:`qsize` of the queue is the sum of the sizes
        of all the items. The default size for an item is 1.
        """
        if size is None:
            size = 1
        with self._lock:
            priority = self._get_item_priority(item)
            while self._size + size > self.maxsize > 0:
                if not block:
                    raise QueueFull
                if not self._notfull.wait_for(lambda: self._size+size <= self.maxsize, timeout):
                    raise QueueFull
            heapq.heappush(self._heap, (priority, size, item))
            self._size += size
            self._unfinished_tasks += 1
            self._notempty.notify()

    def put_nowait(self, item, size=None):
        """"Equivalent of ``put(item, False)``."""
        # Don't don't turn this method into a switchpoint as put() will never
        # switch if block is False. This can be done by calling the function
        # wrapped by the @switchpoint wrapper directly.
        return self.put.__wrapped__(self, item, False, size=size)

    @switchpoint
    def get(self, block=True, timeout=None):
        """Pop an item from the queue.

        If the queue is not empty, an item is returned immediately. Otherwise,
        if *block* is True (the default), wait up to *timeout* seconds for an
        item to become available. If not timeout is provided, then wait
        indefinitely.

        If the queue is empty and *block* is false or a timeout occurs, then
        raise a :class:`QueueEmpty` exception.
        """
        with self._lock:
            while not self._heap:
                if not block:
                    raise QueueEmpty
                if not self._notempty.wait(timeout):
                    raise QueueEmpty
            prio, size, item = heapq.heappop(self._heap)
            self._size -= size
            if 0 <= self._size < self.maxsize:
                self._notfull.notify()
        return item

    def get_nowait(self):
        """"Equivalent of ``get(False)``."""
        # See note in put_nowait()
        return self.get.__wrapped__(self, False)

    def task_done(self):
        """Mark a task as done."""
        with self._lock:
            unfinished = self._unfinished_tasks - 1
            if unfinished < 0:
                raise RuntimeError('task_done() called too many times')
            elif unfinished == 0:
                self._alldone.notify()
            self._unfinished_tasks = unfinished

    def join(self):
        """Wait until all tasks are done."""
        with self._lock:
            while self._unfinished_tasks > 0:
                self._alldone.wait()


class LifoQueue(Queue):
    """A queue with LIFO behavior.

    See :class:`Queue` for a description of the API.
    """

    __slots__ = Queue.__slots__

    def _get_item_priority(self, item):
        # Priority function for a LIFO queue
        self._counter += 1
        return -self._counter


class PriorityQueue(Queue):
    """A priority queue.

    Items that are added via :meth:`~Queue.put` are typically ``(priority,
    item)`` tuples. Lower values for priority indicate a higher priority.

    See :class:`Queue` for a description of the API.
    """

    __slots__ = Queue.__slots__

    def _get_item_priority(self, item):
        # Priority function for a priority queue: item should typically be a
        # (priority, item) tuple
        return item
